# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


def is_pow2(n):
    return (n & (n - 1) == 0) and (n > 0)


def get_hadK(n, transpose=False):
    hadK, K = None, None
    if n % 172 == 0:  # llama-2-7b up (11008 = 172 * 64)
        assert is_pow2(n // 172)
        K = 172
        hadK = get_had172().T if transpose else get_had172()
    elif n % 28 == 0:  # llama-3 up (14336 = 28 * 512)
        assert is_pow2(n // 28)
        K = 28
        hadK = get_had28().T if transpose else get_had28()
    elif n % 108 == 0:  # llama-1-13b intermediate (13392 = 108 * 124)
        assert is_pow2(n // 108)
        K = 108
        hadK = get_had108().T if transpose else get_had108()
    elif n % 20 == 0:  # llama2-13b hidden (5120 = 20 * 256)
        assert is_pow2(n // 20)
        K = 20
        hadK = get_had20().T if transpose else get_had20()
    else:
        assert is_pow2(n)
        K = 1
    return hadK, K


def get_had12():
    return torch.FloatTensor(
        [
            [+1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1],
            [+1, +1, -1, +1, -1, -1, -1, +1, +1, +1, -1, +1],
            [+1, +1, +1, -1, +1, -1, -1, -1, +1, +1, +1, -1],
            [+1, -1, +1, +1, -1, +1, -1, -1, -1, +1, +1, +1],
            [+1, +1, -1, +1, +1, -1, +1, -1, -1, -1, +1, +1],
            [+1, +1, +1, -1, +1, +1, -1, +1, -1, -1, -1, +1],
            [+1, +1, +1, +1, -1, +1, +1, -1, +1, -1, -1, -1],
            [+1, -1, +1, +1, +1, -1, +1, +1, -1, +1, -1, -1],
            [+1, -1, -1, +1, +1, +1, -1, +1, +1, -1, +1, -1],
            [+1, -1, -1, -1, +1, +1, +1, -1, +1, +1, -1, +1],
            [+1, +1, -1, -1, -1, +1, +1, +1, -1, +1, +1, -1],
            [+1, -1, +1, -1, -1, -1, +1, +1, +1, -1, +1, +1],
        ]
    )


def get_had28():
    return torch.FloatTensor(
        [
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
        ]
    )


def get_had40():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
        ]
    )


def get_had20():
    return torch.FloatTensor(
        [
            [+1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1],
            [+1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1],
            [+1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1],
            [+1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1],
            [+1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1],
            [+1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1],
            [+1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1],
            [+1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1],
            [+1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1],
            [+1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1],
            [+1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1, -1],
            [+1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1, +1],
            [+1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1, -1],
            [+1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1, -1],
            [+1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1, -1],
            [+1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1, -1],
            [+1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1, +1],
            [+1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1, +1],
            [+1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1, -1],
            [+1, -1, +1, +1, -1, -1, -1, -1, +1, -1, +1, -1, +1, +1, +1, +1, -1, -1, +1, +1],
        ]
    )


def get_had36():
    return torch.FloatTensor(
        [
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
        ]
    )


def get_had60():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
        ]
    )


def get_had52():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
        ]
    )


def get_had108():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
        ]
    )


def get_had140():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
        ]
    )


def get_had156():
    return torch.FloatTensor(
        [
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
        ]
    )


def get_had172():
    return torch.FloatTensor(
        [
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
            ],
            [
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
            ],
            [
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
            ],
            [
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
            ],
            [
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
            ],
            [
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
            ],
            [
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
            [
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                -1,
            ],
            [
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
            ],
            [
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                -1,
                +1,
                -1,
                -1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                +1,
                +1,
                +1,
                -1,
                +1,
                -1,
                +1,
                +1,
                +1,
                +1,
                -1,
                -1,
                +1,
                +1,
                -1,
                -1,
                -1,
                +1,
            ],
        ]
    )
